package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParser
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.defs.TileDefinitionBuilder
import ru.dbotthepony.kstarbound.world.World
import java.io.File
import java.io.FileNotFoundException

class TileDefLoadingException(message: String, cause: Throwable? = null) : IllegalStateException(message, cause)

object Starbound {
	private val tiles = HashMap<String, TileDefinition>()
	val tilesAccess = object : Map<String, TileDefinition> by tiles {}
	val world = World()

	var initializing = false
		private set
	var initialized = false
		private set

	private val _filepath = ArrayList<File>()
	val filepath = object : List<File> by _filepath {}

	fun addFilePath(path: File) {
		_filepath.add(path)
	}

	fun findFile(path: File): File {
		if (path.exists()) {
			return path.canonicalFile
		}

		for (sPath in _filepath) {
			val newPath = File(sPath.path, path.path)

			if (newPath.exists()) {
				return newPath
			}
		}

		throw FileNotFoundException("Unable to find $path in any of known file paths")
	}

	fun findFile(path: String) = findFile(File(path))

	fun loadJson(path: String): JsonElement {
		if (path[0] == '/')
			return JsonParser.parseReader(findFile(path.substring(1)).bufferedReader())

		return JsonParser.parseReader(findFile(path).bufferedReader())
	}

	fun getTileDefinition(name: String): TileDefinition? {
		return tiles[name]
	}

	private val initCallbacks = ArrayList<() -> Unit>()

	fun initializeGame(callback: (finished: Boolean, replaceStatus: Boolean, status: String) -> Unit) {
		if (initializing) {
			throw IllegalStateException("Already initializing!")
		}

		if (initialized) {
			throw IllegalStateException("Already initialized!")
		}

		initializing = true

		Thread {
			val time = System.currentTimeMillis()
			callback(false, false, "Loading materials...")

			loadTileMaterials {
				if (terminateGame) {
					throw InterruptedException("Game is terminating")
				}

				callback(false, true, it)
			}

			callback(false, true, "Loaded materials")

			initializing = false
			initialized = true
			callback(true, false, "Finished loading in ${System.currentTimeMillis() - time}ms")
		}.start()
	}

	fun onInitialize(callback: () -> Unit) {
		if (initialized) {
			callback()
		} else {
			initCallbacks.add(callback)
		}
	}

	fun pollCallbacks() {
		if (initialized && initCallbacks.isNotEmpty()) {
			for (callback in initCallbacks) {
				callback()
			}

			initCallbacks.clear()
		}
	}

	private fun loadTileMaterials(callback: (String) -> Unit) {
		for (sPath in _filepath) {
			val newPath = File(sPath.path, "tiles/materials")

			if (newPath.exists() && newPath.isDirectory) {
				val findFiles = newPath.listFiles()!!

				for (listedFile in findFiles) {
					if (listedFile.path.endsWith(".material")) {
						try {
							callback("Loading ${listedFile.name}")
							val tileDef = TileDefinitionBuilder.fromJson(JsonParser.parseReader(listedFile.bufferedReader()) as JsonObject).build(listedFile.parent)

							check(tiles[tileDef.materialName] == null) { "Already has material with ID ${tileDef.materialName} loaded!" }
							tiles[tileDef.materialName] = tileDef
						} catch(err: Throwable) {
							throw TileDefLoadingException("Loading tile file ${listedFile.name}", err)
						}
					}
				}
			}
		}
	}
}
